#! /usr/bin/env python2.5
# -*- coding: utf-8 -*-
###########################################################*############################################################
#
#                                                ColorTouch.indigoPlugin
#                                                        1.0.0
#
#                                                an Indigo Server Plugin 
#                                                       for the 
#                                       Venstar ColorTouch Residential Thermostats 
#                                           with the optional SkyPort Wi-Fi Key
#
#                                   Copyright (c) 2015, Rick Neil. All rights reserved.
#
#
###########################################################*############################################################
# Requirements:
#	• Indigo v6+, API 1.17 (http://www.indigodomo.com)
#	• Venstar ColorTouch Thermostat; minimum on-board API: v4 (http://www.indigodomo.com)
#	  Supported ColorTouch models: T5800, T5900 (residential)
#	• Venstar SkyPort Wi-Fi Key (http://venstar.com/Thermostats/ACC0454/)
#	• Wi-Fi network
###########################################################*############################################################
# Notes:
#	• This plugin allows full control and status monitoring of ColorTouch residential Thermostats via Indigo 6.x.
#	• The Venstar SkyPort Wi-Fi Key (P/N: ACC0454, sold separately) adds Wi-Fi capability to ColorTouch
#	  Thermostats and is required.
#	• Venstar's optional SkyPort Cloud online service is NOT required.  This plugin communicates directly with 
#	  the ColorTouch/SkyPort Wi-Fi Key over your local Wi-Fi network; an Internet connection is not required.
#	• All features provided through the ColorTouch API v4 have been implemented, except for auto-discovery
#	  via Simple Service Discovery Protocol (SSDP).
#	• A local network IP address (or dns name which correctly resolves to one) is required when initially 
#	  configuring the plugin.  It is strongly recommended that you configure your ColorTouch with a static IP
#	  address.  Doing so will ensure that the plugin can easily re-establish communication with the ColorTouch
#	  whenever it is interrupted.
###########################################################*############################################################

########################################################################################################################

# #mark Imports
########################################################################################################################

import indigo

# Python 2.5 imports
import os
import sys
import random
try:
	import json
except:
	import simplejson as json
import socket
import urllib
import urllib2
import datetime
import time
import appscript

# additional plugin files
import util
import muddle

########################################################################################################################

# #mark ColorTouch/Indigo HVAC Mode Maps
########################################################################################################################

######################
# ColorTouch reports its HVAC mode in the "mode" property (0=Off, 1=Heat, 2=Cool, 3=Auto); it does not include the Schedule (Program) Mode
# ColorTouch reports its Program Mode in the "schedule" property (0=Off, 1=On)
#	When the schedule is active, the Mode will be either Heat, Cool, or Auto as determined by the programmed schedule
#	There is only one Schedule, and it cannot be modified in any way via the API -- in can only be turned On or Off.
# map ColorTouch main mode and schedule values to indigo.kHvacMode constants
# indigoHvacOperationMode = ctHvacModeMap[ctMode.xxx + 4*ct_schedule]
ctHvacModeMap = [indigo.kHvacMode.Off, indigo.kHvacMode.Heat, indigo.kHvacMode.Cool, indigo.kHvacMode.HeatCool, "n/a", indigo.kHvacMode.ProgramHeat, indigo.kHvacMode.ProgramCool, indigo.kHvacMode.ProgramHeatCool]

# ColorTouch reports its Fan Mode in the "fan" property (0=Auto, 1=On)
# map ColorTouch fan mode values to indigo.kFanMode constants
ctFanModeMap = [indigo.kFanMode.Auto, indigo.kFanMode.AlwaysOn]

########################################################################################################################

# #mark ColorTouch enumerations and constants
########################################################################################################################

######################
# ColorTouch enumerations
ctMode = util.enum(Off=0, Heat=1, Cool=2, Auto=3, none=4, Schedule_Heat=5, Schedule_Cool=6, Schedule_Auto=7)
ctFan = util.enum(Auto=0, On=1)
ctAwayMode = util.enum(Home=0, Away=1)
ctSchedule = util.enum(Off=0, Onboard=1)
ctSchedulePart = util.enum(Morning=0, Day=1, Evening=2, Night=3, Inactive=255)
ctTempUnits = util.enum(Fahrenheit=0, Celsius=1)
ctAvailableModes = util.enum(All=0, HeatAndCool=1, Heat=2, Cool=3)
ctState = util.enum(Off=0, Heating=1, Cooling=2, Lockout=3, Error=4)
ctFanState = util.enum(Off=0, On=1)
ctEnabled = util.enum(Disabled=0, Enabled=1)
ctActive = util.enum(Inactive=0, Active=1)
ctIncrease = util.enum(increase=False, decrease=True)
ctOnOff = util.enum(Off=0, On=1)
ctAlertsPolling = util.enum(Off=0, Status=1, Hourly=3600, Dialy=86400)
######################
# Query (HTTP GET) resources, from the root resource resource (IP address) 
# responses are JSON encoded
ctQueryAPIInfo = "/"
ctQueryInfo = "/query/info"
ctQuerySensors = "/query/sensors"
ctQueryRuntimes = "/query/runtimes"
ctQueryAlerts = "/query/alerts"

######################
# Update (HTTP POST) resources, from the root resource (IP address)
# parameters are URL encoded
ctSendControl = "/control"
ctSendSettings = "/settings"

######################
# Humidify and dehumidify setpoint limits
ctHumMin = 0
ctHumMax = 60
ctDeHumMin = 25
ctDeHumMax = 99
ctHumiditySPDelta = 4

######################
# date format strings
ctDateTimeFormat = "%Y-%m-%d %H:%M:%S"
ctDateTimeFormat2 = "%m/%d/%y, %l:%M %p"

# the sleep time (in seconds) between iterations of the concurrent thread
# a lower number will more aggressively test if the polling interval for each
# device has elapsed since the last state refresh, but at the cost of higher CPU usage
# tweak/balace to your requirements and your system resources
ctConcurrentThreadSleepTime = 0.5

########################################################################################################################
class Plugin(indigo.PluginBase):
########################################################################################################################

# #mark Internal Class Methods
########################################################################################################################

	######################
	# standard PluginBase __init__ method with a bit of private housekeeping
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
		self.debug = False
		self.simulateTempChanges = False		# Every few seconds update to random temperature values
		self.simulateHumidityChanges = False	# Every few seconds update to random humidity values

		# set to False to temporarily disable Auto-Refresh, even when dev.pluginProps['autoRefresh'] is True
		self.honorAutoRefresh = True
		
		# optionally log all state changes as they are updated on the Indigo server
		self.logStateChanges = False
		
		# optionally suppress logging in Action Callback methods (False = suppress logging)
		self.actionlogRefresh = True

		self.pPrefs = pluginPrefs
		
		self.lastRefresh = {}

	######################
	# standard PluginBase __del__ method
	def __del__(self):
		indigo.PluginBase.__del__(self)

########################################################################################################################

# #mark Private Methods
########################################################################################################################
		
	######################
	# Update state on server, with some private housekeeping
	def _ctUpdateStateOnServer(self, dev, theKey, theValue, theUIValue = ""):
		# reduce unnecessary chatter with the server
		if dev.states.has_key(theKey) and dev.states[theKey] != theValue:

			# optionally log state changes (for debugging)
			if self.logStateChanges:
				indigo.server.log("**Updating '%s' to '%s' (was '%s') %s" % (theKey, str(theValue), str(dev.states[theKey]), theUIValue)) 

			# update the state on the server
			dev.updateStateOnServer(theKey, theValue, uiValue = theUIValue)

			# only update the Schedule Mode Description if needed
			if theKey == "ctScheduleMode":
				dev.updateStateOnServer("ctScheduleModeDescription", self._scheduleDescription(dev, theValue))
			if theKey == "ct_schedulepart":
				dev.updateStateOnServer("ctScheduleModeDescription", self._scheduleDescription(dev, dev.states["ct_schedule"]))

			# save the date/time of this update
			rightNow = datetime.datetime.now().strftime(ctDateTimeFormat)
			rightNow2 = datetime.datetime.now().strftime(ctDateTimeFormat2)
			dev.updateStateOnServer("ctLastUpdate", rightNow, uiValue=rightNow2)

	######################
	# Query the hardware for current status and update the Indigo Server with any changes
	def _refreshStatesFromHardware(self, dev, logRefresh, commJustStarted):
		numTemps = int(dev.pluginProps["NumTemperatureInputs"])
		numHums = int(dev.pluginProps["NumHumidityInputs"])

		if logRefresh:
			indigo.server.log(u"sending \"%s\" status request" % dev.name)

		# For debugging, we can just update the temperature and humidity
		# sensor states to random values, without querying the hardware.
		if self.simulateTempChanges:
			# For debugging
			# Simulate changing temperature values coming in from the
			# hardware by updating all temp values randomly:
			for index in range(1, numTemps + 1):
				exampleTemp = random.randint(62, 88)
				stateKey = u"temperatureInput" + str(index)
				self._ctUpdateStateOnServer(dev, stateKey, exampleTemp)
				if logRefresh:
					indigo.server.log(u"received \"%s\" temperature%d update to %.1f°" % (dev.name, index, exampleTemp))
		if self.simulateHumidityChanges:
			# For debugging
			# Simulate changing humidity values coming in from the
			# hardware by updating humidity value randomly (if there is a humidty
			# sensor on this thermostat model - T5900, only):
			for index in range(1, numHums + 1):
				exampleHumidity = random.randint(15, 90)
				self._ctUpdateStateOnServer(dev, "humidityInput" + str(index), exampleHumidity)
				if logRefresh:
					indigo.server.log(u"received \"%s\" humidity%d update to %.0f%%" % (dev.name, index, exampleHumidity))

		#if not debugging; query the hardware
		if not (self.simulateHumidityChanges and self.simulateTempChanges):
			# if the device is configured with two temperature sensors, read
			# them both and update the Indigo server accordingly.  If not,
			# there is no need to aceess the ctQuerySensors resource, as we can
			# safely assume that the "spacetemp" property reported by the 
			# ctQueryInfo resource below, is the correct value for temperature 
			# sensor 1.
			if numTemps == 2:
				response = self._queryTstatHardware(dev, ctQuerySensors)
				if response[0]:
					for sensor in response[1]["sensors"]:
						# the API documentation (http://developer.venstar.com/restcalls.html)
						# says the key should be "Outdoor"; my stat returns "Remote"
						if sensor["name"] in ("Remote", "Outdoor"):
							self._ctUpdateStateOnServer(dev, "temperatureInput2", sensor["temp"])
						elif sensor["name"] == "Thermostat":
							self._ctUpdateStateOnServer(dev, "temperatureInput1", sensor["temp"])

			# access the ctQueryInfo resource to get current states
			response = self._queryTstatHardware(dev, ctQueryInfo)
			if response[0]:
				info = response[1]

				# update the rest of the states defined by the Indigo Thermostat Device class

				# temperature sensors
				if numTemps == 1:
					self._ctUpdateStateOnServer(dev, "temperatureInput1", info["spacetemp"])

				# humidity sensor (if available on hardware)
				if numHums == 1:
					self._ctUpdateStateOnServer(dev, "humidityInput1", info["hum"])

				# temperature setpoints (heat and cool)
				self._ctUpdateStateOnServer(dev, "setpointHeat", info["heattemp"])
				self._ctUpdateStateOnServer(dev, "setpointCool", info["cooltemp"])

				# operating mode -- (Off, Heat, Cool, Auto, Schedule (Heat), Schedule (Cool), Schedule (Auto))
				hvacOpMode = (info["mode"] + 4*info["schedule"])
				self._ctUpdateStateOnServer(dev, "hvacOperationMode", ctHvacModeMap[hvacOpMode])

				# fan mode (On, Auto)
				self._ctUpdateStateOnServer(dev, "hvacFanMode", ctFanModeMap[(info["fan"])])

				# A/C (compressor) state (On, Off)
				self._ctUpdateStateOnServer(dev, "hvacCoolerIsOn", info["state"] == 2)

				# Heater state (On, Off)
				self._ctUpdateStateOnServer(dev, "hvacHeaterIsOn", info["state"] == 1)

				# Fan state (On, Off)
				self._ctUpdateStateOnServer(dev, "hvacFanIsOn", info["fanstate"] == 1)

				# save the current schedule value; if it has changed, we need to do more than simply update the corresponding device state
				savedSchedule = dev.states["ct_schedule"] if dev.states.has_key("ct_schedule") else "-1"

				# update the custom states defined in our Devices.xml file that are 
				# not defined by the default Indigo Thermostat Device Class, but are
				# reported by the hardware when queried
				for key in dev.states.keys():
					if "ct_" in key and key[3:] in info:
						if key == "ct_schedulepart":
							self._ctUpdateStateOnServer(dev, key, info[key[3:]], theUIValue=ctSchedulePart.name[info[key[3:]]])
						else:
							self._ctUpdateStateOnServer(dev, key, info[key[3:]])

				# if the user has turned the Schedule on or off at the thermostat, we need to update more than just dev.states["ct_schedule"]
				newSchedule = info["schedule"]
				if newSchedule != savedSchedule:
					if self.actionlogRefresh:
						indigo.server.log(u"received from \"%s\" set onboard schedule to %s" % (dev.name, ctOnOff.name[newSchedule]))
					self._handleServerScheduleChange(dev, newSchedule)
				
		# optionally, update the log with a detailed status report
		if logRefresh:
			indigo.server.log(u"received \"%s\" status update:" % dev.name)
			indigo.server.log(u"   System Mode: %s | Fan Mode: %s | Heat To: %.1f | Cool To: %.1f" %
			(
				ctMode.name[info["mode"] + 4*info["schedule"]],
				ctFan.name[info["fanstate"]],
				dev.states["setpointHeat"],
				dev.states["setpointCool"]
			))

			indigo.server.log(u"   System State: %s | Fan State: %s | Away Mode: %s | Schedule: %s (%s)" %
			(
				ctState.name[dev.states["ct_state"]],
				ctFanState.name[int(dev.states["hvacFanIsOn"])],
				ctAwayMode.name[dev.states["ct_away"]],
				ctSchedule.name[dev.states["ct_schedule"]],
				ctSchedulePart.name[dev.states["ct_schedulepart"]]
			))

			indigo.server.log(u"   Humidify Setpoint: %d | Dehumidify Setpoint: %d | Available modes: %s" % (dev.states["ct_hum_setpoint"], dev.states["ct_dehum_setpoint"], ctAvailableModes.name[dev.states["ct_availablemodes"]]))

	#####################
	# Process Action request from Indigo Server to change the specified thermostat's main mode, and/or temperature setpoints.
	def _handleChangeHvacModeAction(self, dev, newHvacMode, newHeatTo, newCoolTo):
		if newHvacMode != -1:
			newModeValue = ctHvacModeMap.index(newHvacMode)
			actionStr = ctMode.name[ctHvacModeMap.index(newHvacMode)]
			indigo.server.log(u"sending \"%s\" Mode change to %s (H:%d, C:%d)" % (dev.name, actionStr, int(newHeatTo), int(newCoolTo)))
		else:
			indigo.server.log(u"sending \"%s\" temperature setpoint changes H:%d, C:%d" % (dev.name, int(newHeatTo), int(newCoolTo)))
			newModeValue = -1

		if newHvacMode in (indigo.kHvacMode.Off, indigo.kHvacMode.HeatCool, indigo.kHvacMode.Cool, indigo.kHvacMode.Heat, -1):
			# ensure that the Schedule (Program) is not running, or this mode change command will be ignored
 			theResponse = self._sendToTstatHardware(dev, ctSendSettings, {"schedule" : ctSchedule.Off})
 			if theResponse[0]:
				controlRec = self._modeControlRecord(dev, newModeValue, newHeatTo, newCoolTo)
				if controlRec.has_key("error"):
					theResponse = [False, controlRec["error"]]
				else:
					theResponse = self._sendToTstatHardware(dev, ctSendControl, controlRec)
		# if newHvacMode is one of Indigo's 3 pre-defined HVAC Program Modes, we'll run the ColorTouch's Schedule
		# we can only turn the Schedule On or Off -- we cannot control its HVAC Mode via the API
		elif newHvacMode in (indigo.kHvacMode.ProgramHeat, indigo.kHvacMode.ProgramCool, indigo.kHvacMode.ProgramHeatCool):
			theResponse = self._sendToTstatHardware(dev, ctSendSettings, {"schedule" : ctSchedule.Onboard})
		else:
			theResponse = [False, "unsupported mode"]

		if theResponse[0]:
			# tell the Indigo Server to update
			if newHvacMode != -1:
				self._ctUpdateStateOnServer(dev, "hvacOperationMode", newHvacMode)
			if newHeatTo != -1:
				self._ctUpdateStateOnServer(dev, "setpointHeat", newHeatTo)
			if newCoolTo != -1:
				self._ctUpdateStateOnServer(dev, "setpointCool", newCoolTo)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" Mode change to %s failed [%s]" % (dev.name, actionStr, theResponse[1]), isError=True)


	######################
	# Process Action request from Indigo Server to change thermostat's fan mode.
	def _handleChangeFanModeAction(self, dev, newFanMode):
		# Command hardware module (dev) to change the fan mode here:

		actionStr = ctFan.name[ctFanModeMap.index(newFanMode)]
		indigo.server.log(u"sending \"%s\" Fan Mode change to %s" % (dev.name, actionStr))

		theResponse = self._sendToTstatHardware(dev, ctSendControl, {"fan" : ctFanModeMap.index(newFanMode)})

		if theResponse[0]:
			# tell the Indigo Server to update the state.
			self._ctUpdateStateOnServer(dev, "hvacFanMode", newFanMode)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" Fan Mode change to %s failed" % (dev.name, actionStr), isError=True)

	######################
	# Process Action request from Indigo Server to change a setpoint (heat, cool, humidify, or dehumidify).
	def _handleChangeSetpointAction(self, dev, newSetpoint, logActionName, stateKey):

		indigo.server.log(u"sending \"%s\" %s to %.1f" % (dev.name, logActionName, newSetpoint))

		if stateKey == u"setpointCool":
			# to comply with the ColorTouch API, obtain a control record that includes:
			# 	BOTH setpoints
			# 	and the mode, if the mode is indigo.kHvacMode.HeatCool (Auto)
			paramsRec = self._modeControlRecord(dev, -1, -1, newSetpoint)
			theResource = ctSendControl
		elif stateKey == u"setpointHeat":
			# to comply with the ColorTouch API, obtain a control record that includes:
			# 	BOTH setpoints
			# 	and the mode, if the mode is indigo.kHvacMode.HeatCool (Auto)
			paramsRec = self._modeControlRecord(dev, -1, newSetpoint, -1)
			theResource = ctSendControl
		elif stateKey == u"ct_hum_setpoint":
			# to comply with the ColorTouch API, obtain a settings record that includes
			# 	BOTH setpoints
			paramsRec = self._humSettingsRecord(dev, newSetpoint, -1)
			theResource = ctSendSettings
		elif stateKey == u"ct_dehum_setpoint":
			# to comply with the ColorTouch API, obtain a settings record that includes
			# 	BOTH setpoints
			paramsRec = self._humSettingsRecord(dev, -1, newSetpoint)
			theResource = ctSendSettings

		# Command hardware module (dev) to udpate the set points:
		theResponse = self._sendToTstatHardware(dev, theResource, paramsRec)

		if theResponse[0]:
			if stateKey in (u"setpointCool", u"setpointHeat"):
				# tell the Indigo Server to update the states.
				self._ctUpdateStateOnServer(dev, "setpointCool", paramsRec["cooltemp"])
				self._ctUpdateStateOnServer(dev, "setpointHeat", paramsRec["heattemp"])
			else:
				# tell the Indigo Server to update the states.
				self._ctUpdateStateOnServer(dev, "ct_hum_setpoint", paramsRec["hum_setpoint"])
				self._ctUpdateStateOnServer(dev, "ct_dehum_setpoint", paramsRec["dehum_setpoint"])

			# log that the command was successfully sent.
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" %s to %.1f failed" % (dev.name, logActionName, newSetpoint), isError=True)

	#####################
	# Process Action request from Indigo Server OR from the thermostat to change the current Schedule mode
	# 0 = None (off), 1 = Onboard Schedule, 2 = Schedule Folder #1 (dev.states["ctScheduleFolders"][0]), 3 = Schedule Folder #2 (dev.states["ctScheduleFolders"][1]), etc.
	def _handleServerScheduleChange(self, dev, newSchedMode):
		# Schedule Folders list
		schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])
		theIndex = newSchedMode - 2
		if theIndex >= len(schedFoldersList):
			indigo.server.log(u"\"%s\" invalid Schedule mode: [%d]" % (dev.name, newSchedMode), isError=True)
			return
		elif newSchedMode <= ctSchedule.Onboard:
			folderId = 0
		else:
			folderId = schedFoldersList[theIndex]
			
		# determine the currently running Custom Schedule Folder (if, any)
		if dev.states["ctScheduleMode"] > ctSchedule.Onboard:
			oldFolderId = schedFoldersList[dev.states["ctScheduleMode"] - 2]
		else:
			oldFolderId = 0

		# update Schedule Mode on server
		self._ctUpdateStateOnServer(dev, "ctScheduleMode", newSchedMode)

		# iterate through all Schedule Folders in the Indigo Server database
		for f in indigo.schedules.folders:
			# is this Schedule Folder in our Custom Schedule Folders List?
			if f.id in schedFoldersList:
				# get the Schedule Folder name for the log entry
				folderName = f.name
				# is this the selected Schedule Folder (that is, the one to enable?)
				if f.id == folderId:
					theEnabled = True
					theVerb = "enabling"
				else: # if not, we're going to ensure it's disabled
					theEnabled = False
					theVerb = "disabling"
			
				# post a log entry if this is the current Folder being disabled, or the new folder being enabled; else
				if f.id in (folderId, oldFolderId):
					indigo.server.log(u"\"%s\" %s Schedule Folder '%s' [%d]" % (dev.name, theVerb, folderName, f.id))

				# enable/disable each Indigo Schedule in this Schedule Folder
				for s in indigo.schedules:
					if s.folderId == f.id and s.enabled != theEnabled:
						indigo.schedule.enable(s.id, value=theEnabled)

		if newSchedMode > ctSchedule.Onboard:
			# set the thermostat according to the selected Schedule Folder
			self._currentScheduleForFolder(dev, folderId, True)

	#####################
	# Process Action request from Indigo Server or from this plugin to refresh alert states
	def _handleAlertsRefresh(self, dev, refreshLog, newTime=datetime.datetime.now().strftime(ctDateTimeFormat)):
		if refreshLog:
			indigo.server.log(u"sending \"%s\" Alert Status request" % (dev.name))
		
		response = self._queryTstatHardware(dev, ctQueryAlerts)
		if response[0]:
			activeAlerts = []
			# tell the Indigo Server to update the state.
			for alert in response[1]["alerts"]:
				stateKey = "ctAlert" + alert["name"].replace(" ", "")
				newVal = alert["active"]

				# generate random values for debugging
#				newVal = bool(random.randint(0, 2))

				self._ctUpdateStateOnServer(dev, stateKey, newVal)

				# compile the active alerts for the log entry
				if newVal:
					activeAlerts.append(str(alert["name"]))

			# save the current date and time to the server 
			self._ctUpdateStateOnServer(dev, "ctAlertsUpdated", newTime)

			# log that the command was successfully sent.
			if refreshLog:
				if len(activeAlerts):
					indigo.server.log(u"received \"%s\": Active Alerts: %s" % (dev.name, str(activeAlerts)))
				else:
					indigo.server.log(u"received \"%s\": No active Alerts" % dev.name)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" Alert Status request failed" % dev.name, isError=True)


	#####################
	# Return the Schedule that would have fired most recently had all Schedules in the specified Folder been enabled for the past week; optionally execute the Schedules Actions(s)
	# Note that only Schedules with a dateType of indigo.kDateType.EveryDay, indigo.kDateType.DaysOfWeek, or indigo.kDateType.Absolute will be considered eligible
	# Schedules with other dateTypes cannot currently be tested for eligibility because Schedules are not yet fully implemented in the Indigo Object Model
	def _currentScheduleForFolder(self, dev, folderId, applyIt):
		if folderId not in indigo.schedules.folders:
			indigo.server.log(u"\"%s\" Schedule Folder ID %d not found" % (dev.name, folderId), isError=True)
			return

#		indigo.server.log(u"\"%s\" applying Schedule Folder '%s' [%d]" % (dev.name, indigo.schedules.folders[folderId].name, folderId))

		# get today's date and day of the week
		rightNow = datetime.datetime.now()
		rightNowDow = int(rightNow.strftime("%w"))

		# list of schedules that would have run today had they been enabled before now, and their execution times (plus the the current time)
		unsortedList = [(0, rightNow.time()),]
		schedId = 0

		# iterate through Indigo's Schedules
		for s in indigo.schedules:
			# is this schedule in the desired folder?
			if s.folderId == folderId:
				# is the "days of week" property (accessible via AppleScript, but not python) valid for this Schedule
				if s.dateType in (indigo.kDateType.EveryDay, indigo.kDateType.DaysOfWeek):
					# get the "days of week" property (accessible via AppleScript, but not python)
					sDow = appscript.app('IndigoServer').time_date_actions[s.name].days_of_week.get()
					# is Today in this Schedule's "days of week"
					if sDow[rightNowDow] == '1':
						# add this Schedule to the list of Schedules that would have run today
						unsortedList.append((s.id, s.nextExecution.time()))
				# is the date for this Schedule an absolute date?
				elif s.dateType in (indigo.kDateType.Absolute,):
					# is the absolute date for this Schedule today's date?
					if s.absoluteDate.date() == rightNow.date():
						# add this Schedule to the list of Schedules that would have run today
						unsortedList.append((s.id, s.nextExecution.time()))

		# create a list of eligible Schedules (plus the current time) sorted by execution time; 
		sortedList = sorted(unsortedList, key=lambda tup: tup[1])

		# find the index of the current time in the sorted list
		theIndex = sortedList.index((0, rightNow.time()))
		# if there is an entry in the list before the current time, that's the Schedule we're looking for
		if theIndex > 0:
			schedId = sortedList[theIndex - 1][0]
		# otherwise, repeat the process for each of the last 6 days
		# but this time, we're looking for the Schedule that would have have executed most recently before today
		else:
			# begin with yesterday
			theDate = rightNow - datetime.timedelta(days=1)
			# iterate through the previous six days
			for x in range(6):
				dayOfWeek = int(theDate.strftime("%w"))
				# start with an empty list since we don't need to compare time to right now
				unsortedList = []
				# iterate through Indigo's Schedules
				for s in indigo.schedules:
					# is this schedule in the desired folder?
					if s.folderId == folderId:
						# is the "days of week" property (accessible via AppleScript, but not python) valid for this Schedule
						if s.dateType in (indigo.kDateType.EveryDay, indigo.kDateType.DaysOfWeek):
							# get the "days of week" property (accessible via AppleScript, but not python)
							sDow = appscript.app('IndigoServer').time_date_actions[s.name].days_of_week.get()
							# is this day in the Schedule's "days of week"
							if sDow[dayOfWeek] == '1':
								# add this Schedule to the list of Schedules that would have run today
								unsortedList.append((s.id, s.nextExecution.time()))
						# is the date for this Schedule an absolute date?
						elif s.dateType in (indigo.kDateType.Absolute,):
							# is the absolute date for this Schedule this day's date?
							if s.absoluteDate.date() == theDate.date():
								# add this Schedule to the list of Schedules that would have run today
								unsortedList.append((s.id, s.nextExecution.time()))
				# if there were eligible Schedules found for this day
				if len(unsortedList) > 0:
					# create a list of eligible Schedules sorted by execution time
					sortedList = sorted(unsortedList, key=lambda tup: tup[1])
					# the last Schedule in the sorted list is the Schedule we're looking for
					schedId = sortedList[-1][0]
					break
				# otherwise, set this day to the day before
				theDate = theDate  - datetime.timedelta(days=1)

		if schedId > 0:
			schedName = indigo.schedules[schedId].name
			if applyIt:
				indigo.server.log(u"\"%s\" executing Schedule '%s' [%d]" % (dev.name, schedName, schedId))
				appscript.app('IndigoServer').execute_time_date_action(schedName)
			return schedName
		else:
			indigo.server.log(u"\"%s\" unable to determine schedule to execute" % dev.name)
			return "Unknown"


	#####################
	# query the hardware resource specified for the device (dev)
	# returns list with two elements:
	#	[0]	:	boolean; True if a valid response was received from the hardware, else False
	#	[1]	:	if [0] is True, a record containing the decoded JSON response from the hardware
	#			if [0] is False, an error message string for logging/debugging
	def _queryTstatHardware(self, dev, resource, netAddress=""):
		IPAddress = dev.pluginProps["networkAddress"] if netAddress == "" else netAddress
 		url = "http://" + IPAddress + resource
		success = False
		response = ""

		# set a default timeout so that the plugin doesn't hang for long periods of
		# in the event of a communications failure
		socket.setdefaulttimeout(5)

		try:
			# send the query to the thermostat
			u = urllib2.urlopen(url)
			jsonStr = u.read()
			try:
				# decode the expected JSON response into a python record
				response = json.loads(jsonStr)
				success = True
			except Exception, err:
				# we received a response, but not in the expected JSON format.
				# as such, we can safely assume that the responding device is not
				# a ColorTouch thermostat
				response = "Invalid response: " + str(err)
		except urllib2.HTTPError, err:
			# HTTP error e.g., 404 Not found, 403 Access denied
			# either an invalid resource was specified in the URL or
			# the responding host is not a ColorTouch thermostat
			response = "HTTP Error: " + str(err.code)
		except urllib2.URLError, err:
			# URL Error
			# No response; there is no host responding at the provided URL
			# Either the URL is bad, or the host is down
			response =  "URL Error: " + str(err.reason)

		return(success, response)

	#####################
	# send an HTTP POST request to the hardware at the specified resources, passing the specified parameters
	def _sendToTstatHardware(self, dev, resource, paramsDict):
		IPAddress = dev.pluginProps["networkAddress"]
		url = "http://" + IPAddress + resource
		success = False
		response = ""

		#set a default timeout so that the plugin doesn't hang for long periods of
		#in the event of a communications failure
		socket.setdefaulttimeout(5)

		# if the PIN is set on the thermostat, it must be included in all control/settings POSTs
		if dev.pluginProps["enablePasscode"]:
			devPrefs = json.loads(self.pPrefs["dev"+str(dev.id)])
			paramsDict["pin"] = muddle.demuddle(devPrefs["ctPC"])
			self.pPrefs["dev"+str(dev.id)] = json.dumps(devPrefs)

		# build a URL-encoded control string
		p = urllib.urlencode(paramsDict)
	
		try:
			u = urllib2.urlopen(url, p)
			jsonStr = u.read()
			try:
				# decode the expected JSON response into a python record
				r = json.loads(jsonStr)
				if "success" in r:
					success = True
					response = u"Success"
				else:
					response = r["reason"]
			except Exception, err:
				# we received a response, but not in the expected JSON format.
				# as such, we can safely assume that the responding device is not
				# a ColorTouch thermostat
				response = "Invalid response: " + str(err)
		except urllib2.HTTPError, err:
			# HTTP error e.g., 404 Not found, 403 Access denied
			# either an invalid resource was specified in the URL or
			# the responding host is not a ColorTouch thermostat
			response = "HTTP Error: " + str(err.code)
		except urllib2.URLError, err:
			# URL Error
			# No response; there is no host responding at the provided URL
			# Either the URL is bad, or the host is down
			response =  "URL Error: " + str(err.reason)

		return(success, response)	

	######################
	# build a valid mode-change control record, enforcing these mode-change requirements
	# imposed by the ColorTouch API to avoid an error response from the hardware:
	#	1.	setpointHeat and setpointCool must always be sent together (with or without mode)
	#   2.	if sending mode, both setpoints must also be sent
	#	3.	if mode is Auto (3), setpointCool must be greater than setpointHeat by at least
	#		ct_setpointdelta (which cannot be adjusted via the API)
	#	4.	setpointHeat must be between ct_heattempmin and ct_heattempmax (inclusive)
	#	5.	setpointCool must be between ct_cooltempmin and ct_cooltempmax (inclusive)
	# if any of newMode, newHeatTo, newCoolTo == -1, use the current device state
	def _modeControlRecord(self, dev, newMode, newHeatTo, newCoolTo):
		# get the current state of the hardware
		response = self._queryTstatHardware(dev, ctQueryInfo)

		if not response[0]:
			return {"error":response[1]}

		# get the setpoint limits for this device
		hMin = response[1]["heattempmin"]
		hMax = response[1]["heattempmax"]
		cMin = response[1]["cooltempmin"]
		cMax = response[1]["cooltempmax"]

		# validate the mode
		mode = newMode if newMode != -1 else response[1]["mode"]
		mode = min(mode, ctMode.Auto)
		mode = max(mode, ctMode.Off)

		# validate the heat setpoint
		heatTo = newHeatTo if newHeatTo != -1 else response[1]["heattemp"]
		heatTo = min(heatTo, hMax)
		heatTo = max(heatTo, hMin)

		# validate the cool setpoint
		coolTo = newCoolTo if newCoolTo != -1 else response[1]["cooltemp"]
		coolTo = min(coolTo, cMax)
		coolTo = max(coolTo, cMin)

		# if mode is Auto 
		if mode == ctMode.Auto:
			# get the current setpoint delta for this device
			delta = response[1]["setpointdelta"]

			# if neither setpoint was provided, use existing
			if newHeatTo == newCoolTo == -1:
				coolTo = response[1]["cooltemp"]
				heatTo = response[1]["heattemp"]
				indigo.server.log("\"%s\" using current setpoints. (H: %d, C: %d)" % (dev.name, heatTo, coolTo))

			# are the setpoints appropriately spaced
			elif coolTo < (heatTo + delta):
				if newHeatTo == -1:
					coolTo = max(coolTo, cMin + delta, hMin + delta)
					heatTo = min(heatTo, coolTo - delta)
				elif newCoolTo == -1:
					heatTo = min(heatTo, hMax - delta, cMax - delta)
					coolTo = max(coolTo, heatTo + delta)
					

		# add the required setpoints to the control record
		controlRecord = {"mode" : mode}
		controlRecord["cooltemp"] = coolTo
		controlRecord["heattemp"] = heatTo
		return controlRecord

	######################
	# build a valid humidify/dehmidify-change control record, enforcing these mode-change requirements
	# imposed by the ColorTouch API to avoid an error response from the hardware:
	#	1.	humidify setpoint and dehumidify setpoint must always be sent together (with or without mode; we're doing it without)
	#	2.	humidify setpoint must be between ctHumMin and ctHumMax (inclusive, fixed in thermostat firmware)
	#	3.	dehumidify setpoint must be between ctDeHumMin and ctDeHumMax (inclusive, fixed in thermostat firmware)
	#	4.	the two setpoints must be at least 4 percentage points apart (not documented; determined through experimentation)
	# if either setpoint == -1, use the current device setpoint
	def _humSettingsRecord(self, dev, newHumTo, newDehumTo):
		# get the current state of the hardware
	 	response = self._queryTstatHardware(dev, ctQueryInfo)

		if not response[0]:
			return {"error":response[1]}

		# validate the hum setpoint
		humTo = newHumTo if newHumTo != -1 else response[1]["hum_setpoint"]
		humTo = min(humTo, ctHumMax)
		humTo = max(humTo, ctHumMin)

		# validate the dehum setpoint
		deHumTo = newDehumTo if newDehumTo != -1 else response[1]["dehum_setpoint"]
		deHumTo = min(deHumTo, ctDeHumMax)
		deHumTo = max(deHumTo, ctDeHumMin)

		# are the setpoints appropriately spaced
		if deHumTo < (humTo + ctHumiditySPDelta):
			if newDehumTo == -1:
				humTo = min(humTo, ctHumMax - ctHumiditySPDelta, ctDeHumMax - ctHumiditySPDelta)
				deHumTo = max(deHumTo, humTo + ctHumiditySPDelta)
			else:
				deHumTo = max(deHumTo, ctDeHumMin + ctHumiditySPDelta, ctHumMin + ctHumiditySPDelta)
				humTo = min(humTo, deHumTo - ctHumiditySPDelta)

		# add the required setpoints to the control record
		controlRecord = {}
		controlRecord["dehum_setpoint"] = deHumTo
		controlRecord["hum_setpoint"] = humTo
		return controlRecord


	###############################
	# update the external Runtimes log file with the latest runtime data from the thermostat
	# in response to each query, the thermostat returns a list of runtime data for each of the last seven days, including today
	# each item in the list contains the runtime (in minutes) for each heating, cooling, and aux stage, as well as 
	# Free Cooling.
	# The timestamp for each entry (except today) is midnight UTC, and the runtimes represent the total minutes of operation over 
	# the previous 24 hours
	# For entry dated today, the the data are current as of the time of the query.  As such, today's entry is ignored, and the
	# external log files will therefore be current as of of yesterday upon completion of this method
	def _updateExtRuntimesLog(self, dev, runtimes):

		ctExternalLogsFolder = dev.pluginProps["externalLogsFolder"]
		if ctExternalLogsFolder[-1] != "/" : ctExternalLogsFolder += "/"
		if dev.pluginProps["useExternalLogsSubfolder"]: 
			ctExternalLogsFolder += (dev.name + "/")

		today = datetime.datetime.today()
		header = "Date,Heat1,Heat2,Cool1,Cool2,Aux1,Aux2,FC\n"
	
		# when this method is called during the first week of January, the stat will return runtime data from both this year and last year
		# here, we split the data into separate lists, one for each year to ensure that they are saved to the correct log file (one per year)
		# create an empty dict of log updates; for each k/v pair, the key is the year, the value is a list of entries for that year
		logUpdates = {}
		for entry in runtimes:
			# get the date & time (local time) from this entry
			entryDate = datetime.datetime.strptime(entry['ts'], ctDateTimeFormat)

			entryYear = entryDate.year
			if str(entryYear) not in logUpdates.keys():
				entryList = []
				logUpdates[str(entryYear)] = entryList
			logUpdates[str(entryYear)].append(entry)

		# next, we scan through the dict, and handle each year found separately
		entriesMade = 0 
		logFiles = []
		# iterate through the years (more often than not, there will only be one)
		for theYear in logUpdates.keys():
			# last day of the prior year
			lastDayOfLastYear = "%s-12-31 00:00:00" % (str(int(theYear) - 1))
			lastDate = datetime.datetime.strptime(lastDayOfLastYear, ctDateTimeFormat).date()

			# external log file name, and path
			fileName = "runtimes-%s.csv" % theYear
			filePath = os.path.expanduser(ctExternalLogsFolder + fileName)
		
			# read the log file, and get the date of the last log entry to prevent duplicate entries
			retVal = util.readFromFile(filePath, 'r')
			if retVal[0]:
				previousEntries = retVal[1].split("\n")
				lastLineDate = previousEntries[-2].split(",")[0]
				lastDate = datetime.datetime.strptime(lastLineDate, ctDateTimeFormat).date()

			# next, we iterate through the list of runtime data collected from the thermostat for theYear
			fileData = ""
			for entry in logUpdates[theYear]:
				# update the log file only with newer entries; ignore the entry for today, as it will be incomplete (it won't be complete until midnight tonight)
				# the log will contain only entries for dates before today
				# get the date from this entry, and convert it back to local time (cheat by subtracting 1 second to prevent the date from rolling over)
				entryDate = datetime.datetime.strptime(entry['ts'], ctDateTimeFormat)
				if (entryDate.date() > lastDate) and (entryDate.time().hour == entryDate.time().minute == entryDate.time().second == 0):
					fileData = fileData + ("%s, %d, %d, %d, %d, %d, %d, %d\n" % (entryDate.strftime(ctDateTimeFormat), entry['heat1'], entry['heat2'], entry['cool1'], entry['cool2'], entry['aux1'], entry['aux2'], entry['fc']))
					# keep track of how many log entries we make (and the number of log files updated); it will be reported in the Indigo log when we're finished
					entriesMade += 1
					if theYear not in logFiles:
						logFiles.append(theYear)
			retVal = util.writeToCSVFile(filePath, fileData, "a", header)
			if not retVal[0]:
				indigo.server.log("Error updating Runtimes log file %s [%s]" % (fileName, str(retVal[1])), isError=True)

		if entriesMade > 0:
			entriesText = ("entry", "entries")
			filesText = ("file", "files")
			returnResponse = "%d %s added to %d Runtimes log %s" % (entriesMade, entriesText[int(bool(entriesMade-1))], len(logFiles), filesText[int(bool(len(logFiles)-1))])
		else:
			returnResponse = "No new Runtimes data to log"
		return(True, returnResponse)

	###############################
	# update the external Sensors log file with the current sensor data from the thermostat
	def _updateExtSensorsLog(self, dev, sensorData):
		ctExternalLogsFolder = dev.pluginProps["externalLogsFolder"]
		if ctExternalLogsFolder[-1] != "/" : ctExternalLogsFolder += "/"
		if dev.pluginProps["useExternalLogsSubfolder"]: 
			ctExternalLogsFolder += (dev.name + "/")

		rightNow = datetime.datetime.now()
		theYear = str(rightNow.year)
		theMonth = rightNow.strftime("%m")
		header = "Date,Temp1,Temp2,RoomTemp,Hum\n"
	
		spaceTemp = sensorData["spacetemp"]
		for thisSensor in sensorData["sensors"]:
			if thisSensor["name"] == "Thermostat":
				temp1 = thisSensor["temp"]
				hum = thisSensor["hum"]
			else:
				temp2 = thisSensor["temp"]
			
		thisRecord = "%s,%.1f,%.1f,%.1f,%d\n" % (rightNow.strftime(ctDateTimeFormat), temp1, temp2, spaceTemp, hum)
	
		# external log file name, and path
		fileName = "sensors-%s-%s.csv" % (theYear, theMonth)
		filePath = os.path.expanduser(ctExternalLogsFolder + fileName)
	
		# update the log file (create one, if needed)
		retVal = util.writeToCSVFile(filePath, thisRecord, 'w', header)

		returnResponse = "External Senors log updated"
		if not retVal[0]:
			returnResponse = retVal[1]

		return (retVal[0], returnResponse)

	######################
	# return a description for the specified Schedule Mode
	def _scheduleDescription(self, dev, scheduleMode):
		if scheduleMode == 0:
			return "None (Off)"
		elif scheduleMode == 1:
			return "Onboard [%s]" % ctSchedulePart.name[dev.states["ct_schedulepart"]]
		else:
			# Schedule Folders list
			schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])
			theIndex = scheduleMode - 2
			if (theIndex > (len(schedFoldersList) - 1)):
				indigo.server.log(u"\"%s\" invalid Schedule mode: [%d]" % (dev.name, scheduleMode), isError=True)
				return ""
			else:
				folderId = schedFoldersList[theIndex]
				folderName = indigo.schedules.folders[folderId].name
				schedName = self._currentScheduleForFolder(dev, folderId, False)
				return "%s [%s]" % (folderName, schedName)

################################################################################

# #mark ConfigUI Custom Menu Generators.
################################################################################

	######################
	# generate a menu of available Schedule Modes to choose from (In a Config UI)
	def _schedModeListGenerator(self, filter="", valuesDict=None, typeId="", targetId=0):
		dev = indigo.devices[targetId]
		# the first two items in the menu are always the same
		schedMenu = [("0", "None (Off)"), ("1", "Onboard")]

		# the remaining entries are the Custom Schedule Folders in the Custom Schedule Folders List
		schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])
		i = 2
		for k in schedFoldersList:
			schedMenu.append((str(i), indigo.schedules.folders[k].name))
			i += 1
		return schedMenu

	######################
	# generate a menu of available Schedule Folders to add
	def _addFolderListGenerator(self, filter="", valuesDict=None, typeId="", targetId=0):
		# get the existing list of Schedule Folders
		dev = indigo.devices[targetId]
		schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])

		availableFoldersList = []
		for k in indigo.schedules.folders:
			if k.id not in schedFoldersList:
				availableFoldersList.append((str(k.id), k.name))
		return availableFoldersList
	
	######################
	# generate a menu of available Schedule Folders to remove/reorder
	def _currentFolderListGenerator(self, filter="", valuesDict=None, typeId="", targetId=0):
		# get the existing list of Schedule Folders
		dev = indigo.devices[targetId]
		schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])
		availableFoldersList = []
		i = 0
		for k in schedFoldersList:
			availableFoldersList.append((str(k), "%2d.  %s" % (i, indigo.schedules.folders[k].name)))
			i += 1
		return availableFoldersList

	def _folderPositionListGenerator(self, filter="", valuesDict=None, typeId="", targetId=0):
		# get the existing list of Schedule Folders
		dev = indigo.devices[targetId]
		schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])
		folderPositionList = []
		i = 0
		for k in schedFoldersList:
			folderPositionList.append((i, "%2d.  %s" % (i, indigo.schedules.folders[k].name)))
			i += 1
		return folderPositionList

########################################################################################################################

# #mark Public Methods
########################################################################################################################

	######################
	# standard PluginBase startup() method
	def startup(self):
		self.debugLog(u"startup called")
		self.honorAutoRefresh = True

	######################
	# standard PluginBase shutdown() method
	def shutdown(self):
		self.debugLog(u"shutdown called")

	######################
	# standard PluginBase runConcurrentThread() method
	def runConcurrentThread(self):
		self.debugLog(u"concurrent thread started")
		try:
			while True:
				for dev in indigo.devices.iter("self"):
					if not dev.enabled or not dev.configured:
						continue
						
					# Each thermostat device has it's own polling interval that is
					# user-adjustable and is used to determine how frequently the
					# hardware is actually queried.
					if self.honorAutoRefresh and dev.pluginProps["autoRefresh"]:
						rightNow = datetime.datetime.now()
						
						# determine elapsed time since last query
						thisDevLastRefresh = self.lastRefresh[dev.id] if self.lastRefresh.has_key(dev.id) else (rightNow - datetime.timedelta(days=10))
						elapsedSinceChanged = (rightNow - thisDevLastRefresh)

						# if the elapsed time exceeds or equals the current polling interval for this device, refresh status
						if (elapsedSinceChanged.seconds + elapsedSinceChanged.days * 24 * 3600) >= int(dev.pluginProps["pollingInterval"]):
							self._refreshStatesFromHardware(dev, False, False)
							self.lastRefresh[dev.id] = datetime.datetime.now()
						
						# determine if Alerts need to be refreshed
						alertsInterval = int(dev.pluginProps["alertsInterval"])
						if alertsInterval:
							elapsedSinceAlerted = datetime.timedelta(seconds=10)
							if alertsInterval > 1:
								try:
									lastAlertsRefresh = datetime.datetime.strptime(dev.states["ctAlertsUpdated"], ctDateTimeFormat)
								except:
									lastAlertsRefresh = datetime.datetime.strptime("1776-07-04 12:00:00", ctDateTimeFormat)
								elapsedSinceAlerted = rightNow - lastAlertsRefresh

	 						if (elapsedSinceAlerted.seconds + elapsedSinceAlerted.days * 24 * 3600) >= alertsInterval:
								self._handleAlertsRefresh(dev, False, (rightNow - datetime.timedelta(seconds=ctConcurrentThreadSleepTime)).strftime(ctDateTimeFormat))

				self.sleep(ctConcurrentThreadSleepTime)
		except self.StopThread:
			pass	# Optionally catch the StopThread exception and do any needed cleanup.

	######################
	# standard validateDeviceConfigUi() method
	def validateDeviceConfigUi(self, valuesDict, typeId, devId):
		# verify network connection to thermostat
		IPAddress = valuesDict["networkAddress"]
		indigo.server.log(u"Configuring Indigo device for ColorTouch Model %s @ http://%s..." % (typeId, IPAddress))

 		response = self._queryTstatHardware("", ctQueryAPIInfo, IPAddress)
 		if response[0] and "api_ver" in response[1] and "type" in response[1]:
			indigo.server.log(u"  Network configuration successful.")
		else:
			if "URL Error" in str(response[1]):
				additionalInfo = str(response[1]) + "  No response received."
			else:
				additionalInfo = str(response[1]) + "  The device at this network address does not appear to be a ColorTouch Thermostat.  "
			
			indigo.server.log("  Network configuration failed.  " + additionalInfo)
			errorDict = indigo.Dict()
			errorDict["networkAddress"] = "Network configuration failed.  " + additionalInfo + "\n\nVerify that the network address entered here matches the network (IP) address of your ColorTouch Thermostat."
			errorDict["showAlertText"] = "Network configuration failed.  " + additionalInfo + "\n\nVerify that your thermostat is powered up and that the required SkyPort Wi-Fi Key is installed and properly configured.\n\nNext, ensure that the network address you just entered matches the network (IP) address of the thermostat.\n\nYou can use your web browser to verify that the thermostat is responding at this URL: http://" + valuesDict["networkAddress"] + "\n\nYou should receive a response that looks something like this:\n\n{\"api_ver\":4,\"type\":\"residential\"}"
			return (False, valuesDict, errorDict)

		# verify Passcode (if entered)
		if valuesDict["enablePasscode"]:
			if not (valuesDict["PasscodeEntry1"].isdigit() and len(valuesDict["PasscodeEntry1"]) == 4):
				errorDict = indigo.Dict()
				errorDict["PasscodeEntry1"] = "Invalid Passcode.  The Passcode must be exactly four (4) numeric digits, and must exactly match the Passcode entered via the thermostat's touchscreen in the Security Settings."
				errorDict["showAlertText"] =  "Invalid Passcode.  The Passcode must be exactly four (4) numeric digits, and must exactly match the Passcode entered via the thermostat's touchscreen in the Security Settings."
				return (False, valuesDict, errorDict)
			elif valuesDict["PasscodeEntry1"] != valuesDict["PasscodeEntry2"]:
				errorDict = indigo.Dict()
				errorDict["PasscodeEntry2"] = "Passcodes Do Not Match."
				return (False, valuesDict, errorDict)
		
			# muddle the passcode, and save it
			devPrefs = json.loads(self.pPrefs["dev"+str(devId)])
			devPrefs["ctPC"] = muddle.muddle(valuesDict["PasscodeEntry1"])
			self.pPrefs["dev"+str(devId)] = json.dumps(devPrefs)
			valuesDict["PasscodeEntry1"] = ""
			valuesDict["PasscodeEntry2"] = ""
			indigo.server.log(u"  Passcode saved.")

#  		indigo.server.log("valuesDict:")
#  		for x, y in valuesDict.iteritems():
#  			indigo.server.log("   %s: %s" % (x,  y))

		return (True, valuesDict)

	######################
	# standard Device Plugin deviceStartComm() method
	def deviceStartComm(self, dev):
		# force a refresh of device states from the Device.xml file; allows adding device states to existing
		# ColorTouch thermostat devices by simply updating Devices.xml, and reloading this plugin.
		dev.stateListOrDisplayStateIdChanged()

		self.logStateChanges = False
		self.honorAutoRefresh = True
		newProps = dev.pluginProps
		IPAddress = newProps["networkAddress"]

		# displays the IPAddress in the "Address" field in the Indigo Client Devices List
		newProps["address"] = IPAddress	

		indigo.server.log(u"Initiating comm:")
		indigo.server.log(u"  Network address: %s" % IPAddress)
		indigo.server.log(u"  Auto-Refresh: %s" % (ctEnabled.name[int(newProps["autoRefresh"])]))
		if newProps["autoRefresh"]:
			indigo.server.log(u"    Polling interval: %d seconds" % (int(newProps["pollingInterval"])))

		# alerts polling interval
		indigo.server.log(u"    Alerts polling: %s" % ctAlertsPolling.name[int(newProps["alertsInterval"])])

		# query the thermostat for the current firmware API version
 		response = self._queryTstatHardware(dev, ctQueryAPIInfo)
 		if response[0] and "api_ver" in response[1]:
 			newProps["api_ver"] = int(response[1]["api_ver"])
 			newProps["type"] = response[1]["type"]
			indigo.server.log(u"  Communicating with thermostat.  On-board API: v%d (%s)" % ((int(newProps["api_ver"])), newProps["type"]))
		else:
			indigo.server.log(u"  Communication error: %s" % response[1])
		
		# ensure that this device has an entry in the Plugin Prefs Dictionary
		if not self.pPrefs.has_key("dev"+str(dev.id)):
			self.pPrefs["dev"+str(dev.id)] = "{}"
		
		# if Passcode is not enabled, delete previously saved Passcode (if any)
		devPrefs = json.loads(self.pPrefs["dev"+str(dev.id)])
		if not newProps["enablePasscode"] and devPrefs.has_key("ctPC"):
			indigo.server.log("  Deleting old Passcode")
			del(devPrefs["ctPC"])
		self.pPrefs["dev"+str(dev.id)] = json.dumps(devPrefs)
		
		# update the plugin properties on the server
		dev.replacePluginPropsOnServer(newProps)
		
		# query the thermostat for current System/Thermostat status and update the Indigo server
		self._refreshStatesFromHardware(dev, True, True)
		
		# query the thermostat for current Sensor status and update the Indigo server
		self.updateSensors("Ignore This", dev)
		
		# query the thermostat for current Alert status and update the Indigo server
		self.updateAlerts("Ignore This", dev)
		
		# query the thermostat for current Runtime status and update the Indigo server
		self.updateRuntimes("Ignore This", dev)

########################################

# #mark Main Thermostat Action callback
######################

	######################
	# Main Thermostat Action callback
	def actionControlThermostat(self, action, dev):
		doRefresh = True
		
		# set HVAC system mode
		if action.thermostatAction == indigo.kThermostatAction.SetHvacMode:
			disabledMsg = ""
			if dev.states["ct_away"]:
				disabledMsg = "Away Mode is active"
			elif dev.states["ctScheduleMode"]:
				disabledMsg = "Schedule Mode is '%s'" % dev.states["ctScheduleModeDescription"]

			if disabledMsg != "":
				indigo.server.log(u"\"%s\" Mode change request ignored; %s" % (dev.name, disabledMsg), isError=True)
				return
			self._handleChangeHvacModeAction(dev, action.actionMode, -1, -1)

		# set fan mode
		elif action.thermostatAction == indigo.kThermostatAction.SetFanMode:
			self._handleChangeFanModeAction(dev, action.actionMode)

		# change cool setpoint
		elif action.thermostatAction == indigo.kThermostatAction.SetCoolSetpoint:
			newSetpoint = action.actionValue
			self._handleChangeSetpointAction(dev, newSetpoint, u"change cool setpoint", u"setpointCool")

		# change heat setpoint
		elif action.thermostatAction == indigo.kThermostatAction.SetHeatSetpoint:
			newSetpoint = action.actionValue
			self._handleChangeSetpointAction(dev, newSetpoint, u"change heat setpoint", u"setpointHeat")

		# decrease cool setpoint
		elif action.thermostatAction == indigo.kThermostatAction.DecreaseCoolSetpoint:
			newSetpoint = dev.coolSetpoint - action.actionValue
			self._handleChangeSetpointAction(dev, newSetpoint, u"decrease cool setpoint", u"setpointCool")

		# increase cool setpoint
		elif action.thermostatAction == indigo.kThermostatAction.IncreaseCoolSetpoint:
			newSetpoint = dev.coolSetpoint + action.actionValue
			self._handleChangeSetpointAction(dev, newSetpoint, u"increase cool setpoint", u"setpointCool")

		# decrease heat setpoint
		elif action.thermostatAction == indigo.kThermostatAction.DecreaseHeatSetpoint:
			newSetpoint = dev.heatSetpoint - action.actionValue
			self._handleChangeSetpointAction(dev, newSetpoint, u"decrease heat setpoint", u"setpointHeat")

		# increase heat setpoint
		elif action.thermostatAction == indigo.kThermostatAction.IncreaseHeatSetpoint:
			newSetpoint = dev.heatSetpoint + action.actionValue
			self._handleChangeSetpointAction(dev, newSetpoint, u"increase heat setpoint", u"setpointHeat")

		# refresh status
		elif action.thermostatAction in [indigo.kThermostatAction.RequestStatusAll, indigo.kThermostatAction.RequestMode,
		indigo.kThermostatAction.RequestEquipmentState, indigo.kThermostatAction.RequestTemperatures, indigo.kThermostatAction.RequestHumidities,
		indigo.kThermostatAction.RequestDeadbands, indigo.kThermostatAction.RequestSetpoints]:
			self._refreshStatesFromHardware(dev, True, False)
			doRefresh = False

		if doRefresh:
			self._refreshStatesFromHardware(dev, False, False)


################################################################################

# #mark General Action callback
################################################################################

	######################
	# General Device Action callback (none actually supported by the ColorTouch)
	def actionControlGeneral(self, action, dev):
		###### BEEP ######
		if action.deviceAction == indigo.kDeviceGeneralAction.Beep:
			# Beep the hardware module (dev) here:
			indigo.server.log(u"Beep not supported")

		###### ENERGY UPDATE ######
		elif action.deviceAction == indigo.kDeviceGeneralAction.EnergyUpdate:
			# Request hardware module (dev) for its most recent meter data here:
			indigo.server.log(u"Energy Update not supported")

		###### ENERGY RESET ######
		elif action.deviceAction == indigo.kDeviceGeneralAction.EnergyReset:
			# Request that the hardware module (dev) reset its accumulative energy usage data here:
			indigo.server.log(u"Energy Reset not supported")

		###### STATUS REQUEST ######
		elif action.deviceAction == indigo.kDeviceGeneralAction.RequestStatus:
			self._refreshStatesFromHardware(dev, True, False)


################################################################################

# #mark Custom Plugin Action callbacks (defined in Actions.xml)
################################################################################
	
	######################
	# change setpoint(s) -- warmer
	# performs the same function as the "Warmer" button on the touchscrreen
	# In Auto mode, increases BOTH temperature setpoints by 1
	# In Heat mode, increases the Heat To setpoint by 1
	# In Cool mode, increases the Cool To setpoint by 1
	# in Off mode, does nothing.
	def setpointsWarmer(self, pluginAction, dev):
		theMode = dev.hvacMode

		if theMode not in (indigo.kHvacMode.Heat, indigo.kHvacMode.Cool, indigo.kHvacMode.HeatCool):
			return

		newMode = -1
 		
 		if theMode == indigo.kHvacMode.Heat:
 			newHeatTo = dev.heatSetpoint + 1
 			newCoolTo = dev.coolSetpoint

 		if theMode == indigo.kHvacMode.Cool:
 			newHeatTo = dev.heatSetpoint
 			newCoolTo = dev.coolSetpoint + 1

 		if theMode == indigo.kHvacMode.HeatCool:
 			newHeatTo = dev.heatSetpoint + 1
 			newCoolTo = dev.coolSetpoint + 1

		self._handleChangeHvacModeAction(dev, newMode, newHeatTo, newCoolTo)

		
	######################
	# change setpoint(s) -- cooler
	# performs the same function as the "Cooler" button on the touchscrreen
	# In Auto mode, decreases BOTH temperature setpoints by 1
	# In Heat mode, decreases the Heat To setpoint by 1
	# In Cool mode, decreases the Cool To setpoint by 1
	# in Off mode, does nothing.
	def setpointsCooler(self, pluginAction, dev):
		theMode = dev.hvacMode

		if theMode not in (indigo.kHvacMode.Heat, indigo.kHvacMode.Cool, indigo.kHvacMode.HeatCool):
			return

		newMode = -1
 		
 		if theMode == indigo.kHvacMode.Heat:
 			newHeatTo = dev.heatSetpoint - 1
 			newCoolTo = dev.coolSetpoint

 		if theMode == indigo.kHvacMode.Cool:
 			newHeatTo = dev.heatSetpoint
 			newCoolTo = dev.coolSetpoint - 1

 		if theMode == indigo.kHvacMode.HeatCool:
 			newHeatTo = dev.heatSetpoint - 1
 			newCoolTo = dev.coolSetpoint - 1

		self._handleChangeHvacModeAction(dev, newMode, newHeatTo, newCoolTo)

	
	######################
	# Set Away Mode to Away
	def setAwayModeToAway(self, pluginAction, dev):
		# Command hardware module (dev) to set Away Mode to Away:

		if self.actionlogRefresh:
			indigo.server.log(u"sending \"%s\" set Away Mode to Away" % (dev.name))

		theResponse = self._sendToTstatHardware(dev, ctSendSettings, {"away" : ctAwayMode.Away})

		if theResponse[0]:
			# then tell the Indigo Server to update the state:
			self._ctUpdateStateOnServer(dev, "ct_away", ctAwayMode.Away)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" set Away Mode to Away failed [%s]" % (dev.name, theResponse[1]), isError=True)

	######################
	# Set Away Mode to Home
	def setAwayModeToHome(self, pluginAction, dev):
		# Command hardware module (dev) to set Away Mode to Home:

		if self.actionlogRefresh:
			indigo.server.log(u"sending \"%s\" set Away Mode to Home" % (dev.name))

		theResponse = self._sendToTstatHardware(dev, ctSendSettings, {"away" : ctAwayMode.Home})
		if theResponse[0]:
			# tell the Indigo Server to update the state:
			self._ctUpdateStateOnServer(dev, "ct_away", ctAwayMode.Home)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" set Away Mode to Home failed [%s]" % (dev.name, theResponse[1]), isError=True)

	######################
	# Change Schedule Mode
	# 0 = None (off), 1 = Onboard Schedule, 2 = Schedule Folder #1 (dev.states["ctScheduleFolders"][0]), 3 = Schedule Folder #2 (dev.states["ctScheduleFolders"][1]), etc.
	def changeScheduleMode(self, pluginAction, dev):
		self.honorAutoRefresh = False
		try:
			newSchedMode = int(pluginAction.props["scheduleMode"])
		except:
			indigo.server.log(u"\"%s\" Schedule Mode not specified" % (dev.name), isError=True)
			return

		# Schedule Folders list
		schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])
		theIndex = newSchedMode - 2
		if newSchedMode < 0 or (newSchedMode > 1 and theIndex >= len(schedFoldersList)):
			indigo.server.log(u"\"%s\" invalid Schedule mode: [%d]" % (dev.name, newSchedMode), isError=True)
			return

		# the schedule mode parameter for the hardware (0=Off, 1=On)
		newHardwareSchedMode = ctSchedule.Onboard if newSchedMode == ctSchedule.Onboard else ctSchedule.Off

		if self.actionlogRefresh:
			indigo.server.log(u"sending \"%s\" set onboard schedule to %s" % (dev.name, ctOnOff.name[newHardwareSchedMode]))

		# set the hardware's Schedule Mode
		theResponse = self._sendToTstatHardware(dev, ctSendSettings, {"schedule" : newHardwareSchedMode})
		if theResponse[0]:
			# tell the Indigo Server to update the states:
			self._ctUpdateStateOnServer(dev, "ct_schedule", newHardwareSchedMode)
			self._handleServerScheduleChange(dev, newSchedMode)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" set Schedule to On failed [%s]" % (dev.name, theResponse[1]), isError=True)
		self.honorAutoRefresh = True

	######################
	# Add a Schedule Folder to the Custom Schedule Folders List
	def addScheduleFolder(self, pluginAction, dev):
		try:
			folderId = int(pluginAction.props["schedFolderToAdd"])
		except:
			indigo.server.log(u"\"%s\" Schedule Folder not specified" % (dev.name), isError=True)
			return

		schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])
		
		# check to see if the Schedule Folder exists
		if folderId in indigo.schedules.folders:
			# get the Schedule Folder's name
			folderName = indigo.schedules.folders[folderId].name
			# check to see if the Schedule Folder is already in the list
			if folderId not in schedFoldersList:
				# add the named Schedule Folder to the list
				schedFoldersList.append(folderId)
				# log
				indigo.server.log(u"\"%s\" Schedule Folder '%s' [%d] added to Custom Schedule Folders List" % (dev.name, folderName, folderId))
				# update the modified Schedule Folder list on the Server
				self._ctUpdateStateOnServer(dev, "ctScheduleFolders", json.dumps(schedFoldersList))
		else:
			# Else log failure
			indigo.server.log(u"\"%s\" Schedule Folder ID '%d' not found" % (dev.name, folderId), isError=True)

	######################
	# Remove a Schedule Folder from the Custom Schedule Folders List
	def removeScheduleFolder(self, pluginAction, dev):
		try:
			folderId = int(pluginAction.props["schedFolderToRemove"])
		except:
			indigo.server.log(u"\"%s\" Schedule Folder not specified" % (dev.name), isError=True)
			return

		schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])
		
		# check to see if the Schedule Folder exists
		if folderId in indigo.schedules.folders:
			# get the Schedule Folder's name
			folderName = indigo.schedules.folders[folderId].name
			# check to see if the Schedule Folder is in the list
			if folderId in schedFoldersList:
				# remove the named Schedule Folder from the list
				del(schedFoldersList[schedFoldersList.index(folderId)])
				indigo.server.log(u"\"%s\" Schedule Folder '%s' [%d] removed from Custom Schedule Folders List" % (dev.name, folderName, folderId))
				self._ctUpdateStateOnServer(dev, "ctScheduleFolders", json.dumps(schedFoldersList))
			else:
				# Else log failure
				indigo.server.log(u"\"%s\" Schedule Folder '%s' [%d] not found in Custom Schedule Folders List" % (dev.name, folderName, folderId), isError=True)
		else:
			# Else log failure
			indigo.server.log(u"\"%s\" Schedule Folder ID '%d' not found" % (dev.name, folderId), isError=True)

	######################
	# Move a Schedule Folder within the Custom Schedule Folders List
	def moveScheduleFolder(self, pluginAction, dev):
		try:
			folderId = int(pluginAction.props["schedFolderToMove"])
		except:
			indigo.server.log(u"\"%s\" Schedule Folder not specified" % (dev.name), isError=True)
			return

		try:
			toIndex = int(pluginAction.props["toIndex"])
		except:
			indigo.server.log(u"\"%s\" New list index not specified" % (dev.name), isError=True)
			return

		schedFoldersList = [] if dev.states["ctScheduleFolders"] in ["", "{}"] else json.loads(dev.states["ctScheduleFolders"])
		listLen = len(schedFoldersList)
		fromIndex = schedFoldersList.index(folderId)
		
		# check to see if the Schedule Folder exists
		if folderId in indigo.schedules.folders:
			# get the Schedule Folder's name
			folderName = indigo.schedules.folders[folderId].name
			# check to see if the Schedule Folder is in the list
			if folderId in schedFoldersList:
#				if toIndex >= 0 and toIndex < listLen and toIndex != fromIndex:
				if toIndex >= 0 and toIndex < listLen:
					# move the specified Schedule Folder to the specified postion the list
					schedFoldersList.insert(toIndex, schedFoldersList.pop(fromIndex))
					indigo.server.log(u"\"%s\" Schedule Folder '%s' [%d] moved to Custom Schedule Folders List index %d" % (dev.name, folderName, folderId, toIndex))
					self._ctUpdateStateOnServer(dev, "ctScheduleFolders", json.dumps(schedFoldersList))
				else:
					# Else log failure
					indigo.server.log(u"\"%s\" Schedule Folder '%s' [%d] could not be moved.  Bad index: %d" % (dev.name, folderName, folderId, toIndex), isError=True)
			else:
				# Else log failure
				indigo.server.log(u"\"%s\" Schedule Folder '%s' [%d] not found in Custom Schedule Folders List" % (dev.name, folderName, folderId), isError=True)
		else:
			# Else log failure
			indigo.server.log(u"\"%s\" Schedule Folder ID '%d' not found" % (dev.name, folderId), isError=True)
	
	######################
	# Set Temperature Units to Fahrenheit
	def setTempUnitsToF(self, pluginAction, dev):
		# Command hardware module (dev) to set Temp Units to Fahrenheit:

		if self.actionlogRefresh:
			indigo.server.log(u"sending \"%s\" set Temp Units to Fahrenheit" % (dev.name))
			
		theResponse = self._sendToTstatHardware(dev, ctSendSettings, {"tempunits" : ctTempUnits.Fahrenheit})

		if theResponse[0]:
			# tell the Indigo Server to update the state:
			self._ctUpdateStateOnServer(dev, "ct_tempunits", ctTempUnits.Fahrenheit)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" set Temp Units to Fahrenheit failed [%s]" % (dev.name, theResponse[1]), isError=True)

	######################
	# Set Temperature Units to Celsius
	def setTempUnitsToC(self, pluginAction, dev):
		# Command hardware module (dev) to set Temp Units to Celsius:

		if self.actionlogRefresh:
			indigo.server.log(u"sending \"%s\" set Temp Units to Celsius" % (dev.name))

		theResponse = self._sendToTstatHardware(dev, ctSendSettings, {"tempunits" : ctTempUnits.Celsius})

		if theResponse[0]:
			# tell the Indigo Server to update the state:
			self._ctUpdateStateOnServer(dev, "ct_tempunits", ctTempUnits.Celsius)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" set Temp Units to Celsius failed [%s]" % (dev.name, theResponse[1]), isError=True)

	######################
	# Refresh alert states from hardware
	def updateAlerts(self, pluginAction, dev):
		self._handleAlertsRefresh(dev, self.actionlogRefresh)

	######################
	# Refresh runtime data from hardware
	def updateRuntimes(self, pluginAction, dev):

		if self.actionlogRefresh:
			indigo.server.log(u"sending \"%s\" Runtime Status request" % dev.name)

		response = self._queryTstatHardware(dev, ctQueryRuntimes)
		if response[0]:
			today = datetime.datetime.today()
			# calculate the reverse UTC offset to correct the times from the thermostat.  This is necessary because the hardware
			# is not time-zone aware, and it reports it's timestamps as UTC
			isDST = time.daylight and time.localtime().tm_isdst > 0
			revUTCOffset = (time.altzone if (isDST and dev.pluginProps["honorsDST"]) else time.timezone)

			# iterate through the records
			runtimes = response[1]["runtimes"]
			for entry in runtimes:
				# replace the timestamp value with a formatted date string
				entryDate = datetime.datetime.fromtimestamp(entry['ts'] + revUTCOffset)
				entry['ts'] = entryDate.strftime(ctDateTimeFormat)

			# tell the Indigo Server to update the runtime data.
			self._ctUpdateStateOnServer(dev, "ctRuntimes", json.dumps(runtimes))

			# save the current date and time to the server 
			rightNow = datetime.datetime.now().strftime(ctDateTimeFormat)
			self._ctUpdateStateOnServer(dev, "ctRuntimesUpdated", rightNow)

			# log that the command was successfully sent.
			if self.actionlogRefresh:
				indigo.server.log(u"received \"%s\" Runtime Status updated" % dev.name)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" Runtime Status request failed" % dev.name, isError=True)

	######################
	# Refresh runtime data from hardware, and update the external log file
	def updateRuntimesLog(self, pluginAction, dev):
		self.updateRuntimes(pluginAction, dev)
		if self.actionlogRefresh:
			indigo.server.log(u"\"%s\" updating Runtime Status log" % dev.name)
		response2 = self._updateExtRuntimesLog(dev, json.loads(dev.states["ctRuntimes"]))
		indigo.server.log(response2[1])

	######################
	# Refresh sensor data from hardware
	def updateSensors(self, pluginAction, dev):

		if self.actionlogRefresh:
			indigo.server.log(u"sending \"%s\" Sensor Status request" % dev.name)

		response = self._queryTstatHardware(dev, ctQuerySensors)
		if response[0]:
			# tell the Indigo Server to update the sensor data.
			self._ctUpdateStateOnServer(dev, "ctSensors", json.dumps(response[1]["sensors"]))

			for sensor in response[1]["sensors"]:
				# the ColorTouch API documentation (http://developer.venstar.com/restcalls.html)
				# says the key should be "Outdoor"; my stat returns "Remote"
				if sensor["name"] in ("Remote", "Outdoor"):
					self._ctUpdateStateOnServer(dev, "temperatureInput2", sensor["temp"])
				elif sensor["name"] == "Thermostat":
					self._ctUpdateStateOnServer(dev, "temperatureInput1", sensor["temp"])
					if sensor.has_key("hum"):
						self._ctUpdateStateOnServer(dev, "humidityInput1", sensor["hum"])
						
			# save the current date and time to the server 
			rightNow = datetime.datetime.now().strftime(ctDateTimeFormat)
			self._ctUpdateStateOnServer(dev, "ctSensorsUpdated", rightNow)

			# log that the command was successfully sent.
			if self.actionlogRefresh:
				indigo.server.log(u"received \"%s\" Sensor Status updated" % dev.name)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" Sensor Status request failed" % dev.name, isError=True)

	######################
	# Refresh sensor data from hardware, and update the external log file
	def updateSensorsLog(self, pluginAction, dev):

		if self.actionlogRefresh:
			indigo.server.log(u"sending \"%s\" Sensor Status request" % dev.name)

		response = self._queryTstatHardware(dev, ctQuerySensors)
		if response[0]:
			dataDict = response[1]
	
			# tell the Indigo Server to update the runtime data.
			self._ctUpdateStateOnServer(dev, "ctSensors", json.dumps(response[1]["sensors"]))

			# save the current date and time to the server 
			rightNow = datetime.datetime.now().strftime(ctDateTimeFormat)
			self._ctUpdateStateOnServer(dev, "ctSensorsUpdated", rightNow)

			time.sleep(0.5)

			response2 = self._queryTstatHardware(dev, ctQueryInfo)
			if response2[0]:
				if self.actionlogRefresh:
					indigo.server.log(u"\"%s\" updating Sensor Status log" % dev.name)

				dataDict["spacetemp"] = response2[1]["spacetemp"]
				response3 = self._updateExtSensorsLog(dev, dataDict)
		else:
			# Else log failure but do NOT update state on Indigo Server.
			indigo.server.log(u"send \"%s\" Sensor Status log update failed" % dev.name, isError=True)

	
	######################
 	# Set specific mode and setpoints
	def setSpecificMode(self, pluginAction, dev):
		try:
			newMode = int(pluginAction.props["newMode"])
		except:
			newMode = -1
	
		try:
			newHeatTo = int(pluginAction.props["newHeatTo"])
		except:
			newHeatTo = -1
		
		try:
			newCoolTo = int(pluginAction.props["newCoolTo"])
		except:
			newCoolTo = -1
 			
		# set the mode and heat/cool setpoints
		if not newMode == newHeatTo == newCoolTo == -1:
			self._handleChangeHvacModeAction(dev, newMode, newHeatTo, newCoolTo)

		try:
			newFanMode = int(pluginAction.props["newFanMode"])
		except:
			newFanMode = -1
		
		# set the fan mode
		if newFanMode != -1:
			time.sleep(0.25)
			self._handleChangeFanModeAction(dev, newFanMode)

		try:
			newHumTo = int(pluginAction.props["newHumTo"])
		except:
			newHumTo = -1
 			
		# set the humidify setpoint
		if newHumTo != -1:
			stateKey = u"ct_hum_setpoint"
			newSetpoint = newHumTo
			logActionName = u"change humidify setpoint"
			time.sleep(0.5)
			self._handleChangeSetpointAction(dev, newSetpoint, logActionName, stateKey)
		 
		try:
			newDehumTo = int(pluginAction.props["newDehumTo"])
		except:
			newDehumTo = -1
 			
		# set the dehumidify setpoint
		if newDehumTo != -1:
			stateKey = u"ct_dehum_setpoint"
			newSetpoint = newDehumTo
			logActionName = u"change dehumidify setpoint"
			time.sleep(0.5)
			self._handleChangeSetpointAction(dev, newSetpoint, logActionName, stateKey)

	######################
 	# Change humidify setpoint
 	def changeHumidifySetpoint(self, pluginAction, dev):
		stateKey = u"ct_hum_setpoint"
		
		# 0 = relative (change by), 1 = absolute (change to)
		try:
			theMode = int(pluginAction.props["changeMode"])
		except:
			indigo.server.log(u"\"%s\" Change mode not specified" % (dev.name), isError=True)
			return

		if theMode:
			try:
				newSetpoint = int(pluginAction.props["changeToAmount"])
				logActionName = u"change humidify setpoint"
			except:
				indigo.server.log(u"\"%s\" New humidfy setpoint not specified" % (dev.name), isError=True)
				return
		else:
			try:
				newSetpoint = dev.states[stateKey] + int(pluginAction.props["changeByAmount"])
				logActionName = u"%s humidify setpoint" % (ctIncrease.name[newSetpoint < dev.states[stateKey]])
			except:
				indigo.server.log(u"\"%s\" New humidfy setpoint not specified" % (dev.name), isError=True)
				return

		self._handleChangeSetpointAction(dev, newSetpoint, logActionName, stateKey)

	######################
 	# Change dehumidify setpoint
 	def changeDehumidifySetpoint(self, pluginAction, dev):
		stateKey = u"ct_dehum_setpoint"

		# 0 = relative (change by), 1 = absolute (change to)
		try:
			theMode = int(pluginAction.props["changeMode"])
		except:
			indigo.server.log(u"\"%s\" Change mode not specified" % (dev.name), isError=True)
			return

		if theMode:
			try:
				newSetpoint = int(pluginAction.props["changeToAmount"])
				logActionName = u"change dehumidify setpoint"
			except:
				indigo.server.log(u"\"%s\" New dehumidfy setpoint not specified" % (dev.name), isError=True)
				return
		else:
			try:
				newSetpoint = dev.states[stateKey] + int(pluginAction.props["changeByAmount"])
				logActionName = u"%s dehumidify setpoint" % (ctIncrease.name[newSetpoint < dev.states[stateKey]])
			except:
				indigo.server.log(u"\"%s\" New dehumidfy setpoint not specified" % (dev.name), isError=True)
				return

		self._handleChangeSetpointAction(dev, newSetpoint, logActionName, stateKey)


################################################################################

# #mark Plugin Actions defined in MenuItems.xml.
################################################################################
	def setAwayModeToAwayAll(self):
		for dev in indigo.devices.iter("self"):
			self.setAwayModeToAway("Ignore This", dev)

	def setAwayModeToHomeAll(self):
		for dev in indigo.devices.iter("self"):
			self.setAwayModeToHome("Ignore This", dev)

	def setTempUnitsToFAll(self):
		for dev in indigo.devices.iter("self"):
			self.setTempUnitsToF("Ignore This", dev)

	def setTempUnitsToCAll(self):
		for dev in indigo.devices.iter("self"):
			self.setTempUnitsToC("Ignore This", dev)

	def updateAlertsAll(self):
		for dev in indigo.devices.iter("self"):
			self.updateAlerts("Ignore This", dev)

	def updateRuntimesAll(self):
		for dev in indigo.devices.iter("self"):
			self.updateRuntimes("Ignore This", dev)

	def updateRuntimesLogAll(self):
		for dev in indigo.devices.iter("self"):
			self.updateRuntimesLog("Ignore This", dev)

	def updateSensorsAll(self):
		for dev in indigo.devices.iter("self"):
			self.updateSensors("Ignore This", dev)

	def updateSensorsLogAll(self):
		for dev in indigo.devices.iter("self"):
			self.updateSensorsLog("Ignore This", dev)
